@@ -1,7 +1,9 @@
 from .api import *
 from .models import *
+from .protocol import *
 from .inference import *
 from .configs import Configs
+from .data import PrefetchLoader
 from .modules.heads import HeadBase
 from .modules.heads import HeadConfigs
 from .modules.extractors import ExtractorBase
