@@ -16,7 +16,6 @@ from cftool.ml.utils import EnsemblePattern
 from cfdata.tabular import task_type_type
 from cfdata.tabular import parse_task_type
 from cfdata.tabular import TaskTypes
-from cfdata.tabular import TabularData
 
 from ..dist import Task
 from ..dist import Experiments
@@ -24,6 +23,7 @@ from ..types import data_type
 from ..trainer import Trainer
 from ..trainer import IntermediateResults
 from ..pipeline import Pipeline
+from ..protocol import DataProtocol
 from ..misc.toolkit import to_2d
 
 
@@ -129,6 +129,8 @@ def evaluate(
             if y is not None:
                 y = to_2d(y)
             else:
+                if not isinstance(x, str):
+                    raise ValueError("`x` should be str when `y` is not provided")
                 x, y = data.read_file(x, contains_labels=contains_labels)
                 y = data.transform(x, y).y
             if metrics is None:
@@ -266,7 +268,7 @@ def transform_experiments(experiments: Experiments) -> Dict[str, List[Pipeline]]
 
 
 class RepeatResult(NamedTuple):
-    data: Optional[TabularData]
+    data: Optional[DataProtocol]
     experiments: Optional[Experiments]
     pipelines: Optional[Dict[str, List[Pipeline]]]
     patterns: Optional[Dict[str, List[ModelPattern]]]
