@@ -22,7 +22,6 @@ from cfdata.types import np_float_type
 from cfdata.tabular import task_type_type
 from cfdata.tabular import parse_task_type
 from cfdata.tabular import TaskTypes
-from cfdata.tabular import TabularData
 from optuna.trial import Trial
 
 from .basic import *
@@ -30,6 +29,7 @@ from ..misc.toolkit import *
 from .ensemble import Ensemble
 from ..types import data_type
 from ..pipeline import Pipeline
+from ..protocol import DataProtocol
 
 
 class _TunerResult(NamedTuple):
@@ -98,7 +98,8 @@ class _Tuner(LoggingMixin):
                 )
             if y is not None:
                 read_config["y"] = y
-            tr_data = TabularData(**data_config)
+            data_protocol = kwargs.get("data_protocol", "tabular")
+            tr_data = DataProtocol.make(data_protocol, **data_config)
             tr_data.read(x, **read_config)
             self.has_column_names = tr_data._has_column_names
             y = tr_data.processed.y
