@@ -10,7 +10,6 @@ from torch.nn import Module
 from torch.nn import ModuleDict
 from torch.optim import Optimizer
 from cfdata.tabular import ColumnTypes
-from cfdata.tabular import DataLoader
 from cftool.misc import register_core
 from cftool.misc import LoggingMixin
 
@@ -24,6 +23,7 @@ from ..types import tensor_dict_type
 from ..losses import LossBase
 from ..configs import Configs
 from ..configs import Environment
+from ..protocol import DataLoaderProtocol
 from ..misc.toolkit import to_torch
 from ..modules.heads import HeadBase
 from ..modules.heads import HeadConfigs
@@ -32,6 +32,7 @@ from ..modules.transform import Transform
 from ..modules.transform import Dimensions
 from ..modules.transform import SplitFeatures
 from ..modules.extractors import ExtractorBase
+from ..modules.aggregators import AggregatorBase
 
 model_dict: Dict[str, Type["ModelBase"]] = {}
 
@@ -74,6 +75,11 @@ class PipeConfig(NamedTuple):
             prefix = self.extractor_meta_scope
         return f"{self.transform}_{prefix}_{self.extractor_config}"
 
+    @property
+    def extractor_unique_key(self) -> str:
+        prefix = f"{self.extractor_meta_scope}_{self.extractor}"
+        return f"{self.transform}_{prefix}_{self.extractor_config}"
+
     @property
     def head_config_key(self) -> str:
         if self.head_meta_scope is None:
@@ -89,8 +95,8 @@ class ModelBase(Module, LoggingMixin, metaclass=ABCMeta):
     def __init__(
         self,
         environment: Environment,
-        tr_loader: DataLoader,
-        cv_loader: DataLoader,
+        tr_loader: DataLoaderProtocol,
+        cv_loader: Optional[DataLoaderProtocol],
         tr_weights: Optional[np.ndarray],
         cv_weights: Optional[np.ndarray],
     ):
@@ -134,15 +140,14 @@ class ModelBase(Module, LoggingMixin, metaclass=ABCMeta):
             sorted_indices = [idx for idx in sorted(recognizers) if idx != -1]
             for idx in sorted_indices:
                 recognizer = recognizers[idx]
+                assert recognizer is not None
                 if not recognizer.info.is_valid or idx in ts_indices:
                     excluded += 1
                 elif recognizer.info.column_type is ColumnTypes.NUMERICAL:
                     numerical_columns_mapping[idx] = idx - excluded
                 else:
                     str_idx = str(idx)
-                    categorical_dims.append(
-                        self.tr_data.recognizers[idx].num_unique_values
-                    )
+                    categorical_dims.append(recognizer.num_unique_values)
                     encoding_methods.append(
                         self.encoding_methods.setdefault(
                             str_idx, self.default_encoding_method
@@ -202,6 +207,9 @@ class ModelBase(Module, LoggingMixin, metaclass=ABCMeta):
                     extractor_config,
                     head_config,
                 )
+        aggregator = self.config["aggregator"]
+        aggregator_config = self.config["aggregator_config"]
+        self.aggregator = AggregatorBase.make(aggregator, **aggregator_config)
         # caches
         self._transform_cache: Dict[str, Tensor] = {}
         self._extractor_cache: Dict[str, Tensor] = {}
@@ -259,16 +267,17 @@ class ModelBase(Module, LoggingMixin, metaclass=ABCMeta):
             transform = Transform(self.dimensions, **transform_cfg.pop())
         # extractor
         extractor_cfg_key = pipe_config.extractor_config_key
-        extractor_exists = extractor_cfg_key in self.extractors
+        extractor_unique_key = pipe_config.extractor_unique_key
+        extractor_exists = extractor_unique_key in self.extractors
         if pipe_config.reuse_extractor and extractor_exists:
             extractor = self.extractors[extractor_cfg_key]
         else:
             if extractor_exists:
                 new_index = 1
-                new_extractor_cfg_key = extractor_cfg_key
-                while new_extractor_cfg_key in self.extractors:
-                    new_extractor_cfg_key = f"{extractor_cfg_key}_{new_index}"
-                extractor_cfg_key = new_extractor_cfg_key
+                new_extractor_unique_key = extractor_unique_key
+                while new_extractor_unique_key in self.extractors:
+                    new_extractor_unique_key = f"{extractor_cfg_key}_{new_index}"
+                extractor_unique_key = new_extractor_unique_key
             if extractor_cfg_key in self._extractor_configs:
                 extractor_config = self._extractor_configs[extractor_cfg_key]
             else:
@@ -311,10 +320,10 @@ class ModelBase(Module, LoggingMixin, metaclass=ABCMeta):
         head_cfg.inject_dimensions(head_config)
         head = HeadBase.make(pipe_config.head, head_config)
         # gather
-        self.pipes[key] = transform_key, extractor_cfg_key, head_cfg_key
+        self.pipes[key] = transform_key, extractor_unique_key, head_cfg_key
         if not transform_exists:
             self.transforms[transform_key] = transform
-        self.extractors[extractor_cfg_key] = extractor
+        self.extractors[extractor_unique_key] = extractor
         self.heads[key] = head
         # bypass
         if transform.out_dim == 0 or extractor.out_dim == 0:
@@ -351,23 +360,6 @@ class ModelBase(Module, LoggingMixin, metaclass=ABCMeta):
         )
         self.loss_config["input_logits"] = not self.output_probabilities
 
-    def merge_outputs(
-        self,
-        outputs: tensor_dict_type,
-        **kwargs: Any,
-    ) -> tensor_dict_type:
-        # requires returning `predictions` key
-        values = list(outputs.values())
-        output = None
-        for value in values:
-            if value is None:
-                continue
-            if output is None:
-                output = value
-            else:
-                output = output + value
-        return {"predictions": output}
-
     def forward(
         self,
         batch: tensor_dict_type,
@@ -380,7 +372,7 @@ class ModelBase(Module, LoggingMixin, metaclass=ABCMeta):
         x_batch = batch["x_batch"]
         split = self._split_features(x_batch, batch_indices, loader_name)
         outputs = self.execute(split)
-        return self.merge_outputs(outputs, **kwargs)
+        return self.aggregator.reduce(outputs, **kwargs)
 
     def loss_function(
         self,
@@ -468,7 +460,7 @@ class ModelBase(Module, LoggingMixin, metaclass=ABCMeta):
         head_kwargs_dict: Optional[Dict[str, Dict[str, Any]]] = None,
     ) -> tensor_dict_type:
         results: Dict[str, Tensor] = {}
-        for key, (transform_key, executor_key, _) in self.pipes.items():
+        for key, (transform_key, extractor_key, _) in self.pipes.items():
             if key in self.bypassed_pipes:
                 continue
             # transform
@@ -480,16 +472,16 @@ class ModelBase(Module, LoggingMixin, metaclass=ABCMeta):
             # extract
             if extract_kwargs_dict is None:
                 extract_kwargs_dict = {}
-            extracted = self._extractor_cache.get(executor_key)
+            extracted = self._extractor_cache.get(extractor_key)
             if extracted is None:
-                extractor = self.extractors[executor_key]
-                extract_kwargs = extract_kwargs_dict.get(executor_key, {})
+                extractor = self.extractors[extractor_key]
+                extract_kwargs = extract_kwargs_dict.get(extractor_key, {})
                 extracted = extractor(transformed, **extract_kwargs)
                 extracted_shape = extracted.shape
                 if extractor.flatten_ts:
                     if len(extracted_shape) == 3:
                         extracted = extracted.view(extracted_shape[0], -1)
-                self._extractor_cache[executor_key] = extracted
+                self._extractor_cache[extractor_key] = extracted
             # execute
             if head_kwargs_dict is None:
                 head_kwargs_dict = {}
@@ -504,19 +496,6 @@ class ModelBase(Module, LoggingMixin, metaclass=ABCMeta):
         self._transform_cache = {}
         self._extractor_cache = {}
 
-    def _optimizer_step(
-        self,
-        optimizers: Dict[str, Optimizer],
-        grad_scalar: Optional["amp.GradScaler"],  # type: ignore
-    ) -> None:
-        for opt in optimizers.values():
-            if grad_scalar is None:
-                opt.step()
-            else:
-                grad_scalar.step(opt)
-                grad_scalar.update()
-            opt.zero_grad()
-
     def get_split(self, processed: np.ndarray, device: torch.device) -> SplitFeatures:
         with torch.no_grad():
             return self._split_features(to_torch(processed).to(device), None, None)
