@@ -2,12 +2,13 @@ import os
 import torch
 
 import numpy as np
+import matplotlib.pyplot as plt
+import matplotlib.patches as mpatch
 
 from typing import *
-from cfdata.tabular import DataLoader
-from cfdata.tabular import TabularData
 from cfdata.tabular import TabularDataset
 from cftool.ml import ModelPattern
+from cftool.misc import show_or_save
 from cftool.misc import shallow_copy_dict
 from cftool.misc import lock_manager
 from cftool.misc import timing_context
@@ -19,10 +20,13 @@ try:
 except:
     amp = None
 
+from .data import PrefetchLoader
 from .types import data_type
 from .configs import Elements
 from .configs import Environment
 from .trainer import Trainer
+from .protocol import DataProtocol
+from .protocol import DataLoaderProtocol
 from .inference import Inference
 from .inference import PreProcessor
 from .misc.toolkit import to_2d
@@ -31,8 +35,18 @@ from .models.base import model_dict
 from .models.base import ModelBase
 
 
+key_type = Tuple[Union[str, Optional[str]], ...]
+
+
 class Pipeline(LoggingMixin):
     def __init__(self, environment: Environment):
+        # typing
+        self.tr_data: DataProtocol
+        self.cv_data: Optional[DataProtocol]
+        self.tr_loader: DataLoaderProtocol
+        self.tr_loader_copy: DataLoaderProtocol
+        self.cv_loader: Optional[DataLoaderProtocol]
+        # common
         self.environment = environment
         self.device = environment.device
         self.model: Optional[ModelBase] = None
@@ -55,7 +69,7 @@ class Pipeline(LoggingMixin):
     __repr__ = __str__
 
     @property
-    def data(self) -> TabularData:
+    def data(self) -> DataProtocol:
         return self._original_data
 
     @property
@@ -99,13 +113,19 @@ class Pipeline(LoggingMixin):
                 self.ts_label_collator_config,
             )
         self.sampler_config.setdefault("verbose_level", self.data._verbose_level)
-        self.preprocessor = PreProcessor(self._original_data, self.sampler_config)
+        self.preprocessor = PreProcessor(
+            self._original_data,
+            self.loader_protocol,
+            self.sampler_protocol,
+            self.sampler_config,
+        )
         tr_sampler = self.preprocessor.make_sampler(
             self.tr_data,
             self.shuffle_tr,
             self.tr_weights,
         )
-        self.tr_loader = DataLoader(
+        self.tr_loader = DataLoaderProtocol.make(
+            self.loader_protocol,
             self.batch_size,
             tr_sampler,
             return_indices=True,
@@ -116,7 +136,8 @@ class Pipeline(LoggingMixin):
             self.cv_loader = None
         else:
             cv_sampler = self.preprocessor.make_sampler(self.cv_data, False)
-            self.cv_loader = DataLoader(
+            self.cv_loader = DataLoaderProtocol.make(
+                self.loader_protocol,
                 self.cv_batch_size,
                 cv_sampler,
                 return_indices=True,
@@ -147,7 +168,6 @@ class Pipeline(LoggingMixin):
                 raise ValueError(msg)
             self.inference = Inference(
                 self.preprocessor,
-                self.device,
                 model=self.model,
                 binary_config=self.binary_config,
                 use_binary_threshold=self.use_binary_threshold,
@@ -179,7 +199,7 @@ class Pipeline(LoggingMixin):
             self.sample_weights = None
         else:
             self.sample_weights = sample_weights.copy()
-        self._original_data = TabularData(**self.data_config)
+        self._original_data = DataProtocol.make(self.data_protocol, **self.data_config)
         self._original_data.read(*args, **self.read_config)
         self.tr_data = self._original_data
         self._save_original_data = x_cv is None
@@ -227,8 +247,215 @@ class Pipeline(LoggingMixin):
         # logging
         self.log_timing()
 
+    @staticmethod
+    def _rectangle(
+        ax: Any,
+        x: float,
+        y: float,
+        width: float,
+        height: float,
+        color: Any,
+        text: str,
+    ) -> Tuple[float, float]:
+        rectangle = mpatch.Rectangle(
+            (x, y),
+            width,
+            height,
+            color=color,
+            alpha=0.8,
+            ec="#000000",
+        )
+        ax.add_artist(rectangle)
+        rx, ry = rectangle.get_xy()
+        cx = rx + 0.5 * rectangle.get_width()
+        cy = ry + 0.5 * rectangle.get_height()
+        ax.annotate(
+            text,
+            (cx, cy),
+            color="black",
+            fontsize=16,
+            ha="center",
+            va="center",
+        )
+        return cx, cy
+
+    @staticmethod
+    def _arrow(
+        ax: Any,
+        lx: float,
+        ly: float,
+        rx: float,
+        ry: float,
+        half_box_width: float,
+    ) -> None:
+        lx += half_box_width
+        rx -= half_box_width
+        ax.annotate(
+            text="",
+            xy=(rx, ry),
+            xytext=(lx, ly),
+            xycoords="data",
+            arrowprops=dict(arrowstyle="->", color="black"),
+        )
+
+    @staticmethod
+    def _box_msg(key: key_type, delim: str) -> str:
+        scope, meta, config = key[1:]
+        if meta is None:
+            scope_str = ""
+            extractor_str = f"{scope}_{config}"
+        else:
+            scope_str = f"\n{scope}"
+            extractor_str = f"{meta}_{config}"
+        return f"Extractor\n{delim}\n{extractor_str}{scope_str}"
+
+    @staticmethod
+    def _rectangles(
+        ax: Any,
+        color: Any,
+        x: float,
+        y_max: float,
+        box_width: float,
+        box_height: float,
+        delim: str,
+        keys: List[key_type],
+        positions: Dict[key_type, Tuple[float, float]],
+    ) -> None:
+        for i, key in enumerate(keys):
+            y = (i + 0.5) * (y_max / len(keys))
+            args = ax, x, y, box_width, box_height, color
+            cx, cy = Pipeline._rectangle(*args, Pipeline._box_msg(key, delim))
+            positions[key] = cx, cy
+
     # api
 
+    def draw(
+        self,
+        export_path: Optional[str] = None,
+        *,
+        transparent: bool = True,
+    ) -> "Pipeline":
+        pipes = model_dict[self.model_type].registered_pipes
+        if pipes is None:
+            raise ValueError("pipes have not yet been registered")
+        transforms_mapping: Dict[str, str] = {}
+        extractors_mapping: Dict[str, key_type] = {}
+        heads_mapping: Dict[str, key_type] = {}
+        sorted_keys = sorted(pipes)
+        for key in sorted_keys:
+            pipe_cfg = pipes[key]
+            transforms_mapping[key] = pipe_cfg.transform
+            extractor_key: key_type = (
+                pipe_cfg.transform,
+                pipe_cfg.extractor,
+                pipe_cfg.extractor_meta_scope,
+                pipe_cfg.extractor_config,
+            )
+            if not pipe_cfg.reuse_extractor:
+                cursor = 0
+                new_extractor_key: key_type = extractor_key
+                while new_extractor_key in extractors_mapping.values():
+                    cursor += 1
+                    new_extractor_key = extractor_key + (str(cursor),)
+                extractor_key = new_extractor_key
+            extractors_mapping[key] = extractor_key
+            heads_mapping[key] = (
+                key,
+                pipe_cfg.head,
+                pipe_cfg.head_meta_scope,
+                pipe_cfg.head_config,
+            )
+        unique_transforms = sorted(set(transforms_mapping.values()))
+        unique_extractors = sorted(set(extractors_mapping.values()))
+        all_heads = sorted(heads_mapping.values())
+
+        box_width = 0.5
+        box_height = 0.4
+        half_box_width = 0.5 * box_width
+        x_scale, y_scale = 6, 5
+        x_positions = [0, 0.75, 1.5, 2.25]
+        y_gap = box_height * 2.5
+        x_min, x_max = x_positions[0], x_positions[-1]
+        x_diff = x_max - x_min
+        nodes_list = [unique_transforms, unique_extractors, all_heads]
+        y_max = float(max(map(len, nodes_list))) * y_gap  # type: ignore
+        y_max = max(1.5, y_max)
+        fig = plt.figure(dpi=100, figsize=[(x_diff + 2.0) * x_scale, y_max * y_scale])
+
+        ax = fig.add_subplot(111)
+        if transparent:
+            fig.patch.set_alpha(0)
+            ax.patch.set_alpha(0)
+        ax.tick_params(labelbottom=False, bottom=False)
+        ax.tick_params(labelleft=False, left=False)
+        ax.spines["right"].set_visible(False)
+        ax.spines["top"].set_visible(False)
+        ax.spines["left"].set_visible(False)
+        ax.spines["bottom"].set_visible(False)
+        colors = plt.cm.Paired([i / 6 for i in range(4)])
+
+        # rectangle
+        delim = "-" * 16
+        color = colors[0]
+        x = x_positions[0]
+        transform_positions = {}
+        extractor_positions: Dict[key_type, Tuple[float, float]] = {}
+        head_positions: Dict[key_type, Tuple[float, float]] = {}
+        for i, transform in enumerate(unique_transforms):
+            y = (i + 0.5) * (y_max / len(unique_transforms))
+            args = ax, x, y, box_width, box_height, color
+            cx, cy = self._rectangle(*args, f"Transform\n{delim}\n{transform}")
+            transform_positions[transform] = cx, cy
+        color = colors[1]
+        x = x_positions[1]
+        self._rectangles(
+            ax,
+            color,
+            x,
+            y_max,
+            box_width,
+            box_height,
+            delim,
+            unique_extractors,
+            extractor_positions,
+        )
+        color = colors[2]
+        x = x_positions[2]
+        self._rectangles(
+            ax,
+            color,
+            x,
+            y_max,
+            box_width,
+            box_height,
+            delim,
+            all_heads,
+            head_positions,
+        )
+        color = colors[3]
+        x, y = x_positions[3], 0.5 * y_max
+        args = ax, x, y, box_width, box_height, color
+        aggregator = self.environment.model_config["aggregator"]
+        cx, cy = self._rectangle(*args, f"Aggregator\n{delim}\n{aggregator}")
+        aggregator_position = cx, cy
+
+        # arrows
+        for key in sorted_keys:
+            transform = transforms_mapping[key]
+            extractor = extractors_mapping[key]
+            head_tuple = heads_mapping[key]
+            x1, y1 = transform_positions[transform]
+            x2, y2 = extractor_positions[extractor]
+            x3, y3 = head_positions[head_tuple]
+            self._arrow(ax, x1, y1, x2, y2, half_box_width)
+            self._arrow(ax, x2, y2, x3, y3, half_box_width)
+            self._arrow(ax, x3, y3, *aggregator_position, half_box_width)
+
+        ax.set_xlim(x_min, x_max + box_width + 0.1)
+        ax.set_ylim(0, y_max + box_height)
+        show_or_save(export_path, fig)
+        return self
+
     def fit(
         self,
         x: data_type,
@@ -252,9 +479,13 @@ class Pipeline(LoggingMixin):
         returns_probabilities: bool = False,
         **kwargs: Any,
     ) -> Union[np.ndarray, Dict[str, np.ndarray]]:
+        if self.inference is None:
+            raise ValueError("`inference` is not yet generated")
         loader = self.preprocessor.make_inference_loader(
             x,
+            self.device,
             self.cv_batch_size,
+            is_onnx=self.inference.onnx is not None,
             contains_labels=contains_labels,
         )
         kwargs = shallow_copy_dict(kwargs)
@@ -342,6 +573,7 @@ class Pipeline(LoggingMixin):
                 sw_file = os.path.join(data_folder, self.sample_weights_file)
                 np.save(sw_file, self.sample_weights)
             if not self._save_original_data:
+                assert self.cv_data is not None
                 train_data_folder = os.path.join(data_folder, self.train_folder)
                 valid_data_folder = os.path.join(data_folder, self.valid_folder)
                 self.tr_data.save(
@@ -402,13 +634,15 @@ class Pipeline(LoggingMixin):
                 if os.path.isfile(sw_file):
                     sample_weights = np.load(sw_file)
                 # data
+                cv_data: Optional[DataProtocol]
+                data_base = DataProtocol.get(pipeline.data_protocol)
                 original_data_folder = os.path.join(data_folder, cls.original_folder)
                 if not os.path.isdir(original_data_folder):
                     train_data_folder = os.path.join(data_folder, cls.train_folder)
                     valid_data_folder = os.path.join(data_folder, cls.valid_folder)
                     try:
-                        tr_data = TabularData.load(train_data_folder, compress=False)
-                        cv_data = TabularData.load(valid_data_folder, compress=False)
+                        tr_data = data_base.load(train_data_folder, compress=False)
+                        cv_data = data_base.load(valid_data_folder, compress=False)
                     except Exception as e:
                         raise ValueError(
                             f"data information is corrupted ({e}), "
@@ -419,7 +653,7 @@ class Pipeline(LoggingMixin):
                         tr_weights = sample_weights[: len(tr_data)]
                         cv_weights = sample_weights[len(tr_data) :]
                 else:
-                    original_data = TabularData.load(
+                    original_data = data_base.load(
                         original_data_folder,
                         compress=False,
                     )
@@ -447,8 +681,12 @@ class Pipeline(LoggingMixin):
                 pipeline._init_data()
                 pipeline._prepare_modules(is_loading=True)
                 trainer = pipeline.trainer
-                trainer.tr_loader = pipeline.tr_loader
-                trainer.cv_loader = pipeline.cv_loader
+                trainer.tr_loader = PrefetchLoader(pipeline.tr_loader, pipeline.device)
+                cv_loader = pipeline.cv_loader
+                if cv_loader is None:
+                    trainer.cv_loader = None
+                else:
+                    trainer.cv_loader = PrefetchLoader(cv_loader, pipeline.device)
                 trainer.restore_checkpoint(export_folder)
                 trainer._init_metrics()
         return pipeline
