@@ -25,7 +25,6 @@ from cftool.misc import timing_context
 from cftool.misc import Saving
 from cftool.misc import Incrementer
 from cftool.misc import LoggingMixin
-from cfdata.tabular import DataLoader
 
 try:
     amp: Optional[Any] = torch.cuda.amp
@@ -33,9 +32,11 @@ except:
     amp = None
 
 from .misc.toolkit import *
+from .data import PrefetchLoader
 from .configs import Environment
 from .modules import optimizer_dict
 from .modules import scheduler_dict
+from .protocol import DataLoaderProtocol
 from .inference import Inference
 from .models.base import ModelBase
 
@@ -74,7 +75,7 @@ class TrainerState:
         self.max_step_per_snapshot = int(max_step_per_snapshot)
         self.plateau_start = int(plateau_start)
 
-    def inject_loader(self, loader: DataLoader) -> None:
+    def inject_loader(self, loader: DataLoaderProtocol) -> None:
         self.batch_size = loader.batch_size
         self.num_step_per_epoch = len(loader)
 
@@ -439,7 +440,7 @@ class Trainer(MonitoredMixin):
         else:
             self.tracker = Tracker(**environment.tracker_config)
         self.checkpoint_scores: Dict[str, float] = {}
-        self.tr_loader_copy: Optional[DataLoader] = None
+        self.tr_loader_copy: Optional[PrefetchLoader] = None
         self.final_results: Optional[IntermediateResults] = None
         self._use_grad_in_predict = False
         self.onnx: Optional[Any] = None
@@ -448,7 +449,7 @@ class Trainer(MonitoredMixin):
         self.config = environment.trainer_config
         self._verbose_level = environment.verbose_level
         self.update_bt_runtime = self.update_binary_threshold_at_runtime
-        self.scaler = None if amp is None or not self.use_amp else amp.GradScaler()
+        self.grad_scaler = None if amp is None or not self.use_amp else amp.GradScaler()
         self.state = TrainerState(self.config)
         Saving.prepare_folder(self, self.checkpoint_folder)
 
@@ -594,8 +595,10 @@ class Trainer(MonitoredMixin):
             self.metrics_weights.setdefault(metric_type, 1.0)
 
     @property
-    def validation_loader(self) -> DataLoader:
+    def validation_loader(self) -> PrefetchLoader:
         if self.cv_loader is None:
+            if self.tr_loader_copy is None:
+                raise ValueError("`tr_loader_copy` is not yet generated")
             return self.tr_loader_copy
         return self.cv_loader
 
@@ -605,9 +608,11 @@ class Trainer(MonitoredMixin):
         return "tr" if loader is self.tr_loader_copy else "cv"
 
     @property
-    def binary_threshold_loader(self) -> DataLoader:
+    def binary_threshold_loader(self) -> PrefetchLoader:
         if self.cv_loader is not None and len(self.cv_loader.data) >= 1000:
             return self.cv_loader
+        if self.tr_loader_copy is None:
+            raise ValueError("`tr_loader_copy` is not yet generated")
         return self.tr_loader_copy
 
     @property
@@ -623,7 +628,13 @@ class Trainer(MonitoredMixin):
         )
 
     def _optimizer_step(self) -> None:
-        self.model._optimizer_step(self.optimizers, self.scaler)
+        for opt in self.optimizers.values():
+            if self.grad_scaler is None:
+                opt.step()
+            else:
+                self.grad_scaler.step(opt)
+                self.grad_scaler.update()
+            opt.zero_grad()
 
     @staticmethod
     def _metric_verbose(k: str, intermediate: IntermediateResults) -> str:
@@ -729,8 +740,7 @@ class Trainer(MonitoredMixin):
         loss_values = None
         if self._metrics_need_loss:
             loss_dicts = []
-            for (x_batch, y_batch), batch_indices in loader:
-                batch = self.inference.collate_batch(x_batch, y_batch)
+            for batch, batch_indices in loader:
                 with eval_context(self.model):
                     loss_dicts.append(
                         self.model.loss_function(
@@ -809,15 +819,19 @@ class Trainer(MonitoredMixin):
 
     def fit(
         self,
-        tr_loader: DataLoader,
-        tr_loader_copy: DataLoader,
-        cv_loader: DataLoader,
+        tr_loader: DataLoaderProtocol,
+        tr_loader_copy: DataLoaderProtocol,
+        cv_loader: Optional[DataLoaderProtocol],
         tr_weights: Optional[np.ndarray],
         cv_weights: Optional[np.ndarray],
     ) -> None:
-        self.tr_loader = tr_loader
-        self.tr_loader_copy = tr_loader_copy
-        self.cv_loader = cv_loader
+        self.tr_loader = PrefetchLoader(tr_loader, self.device)
+        self.tr_loader_copy = PrefetchLoader(tr_loader_copy, self.device)
+        self.cv_loader: Optional[PrefetchLoader]
+        if cv_loader is None:
+            self.cv_loader = None
+        else:
+            self.cv_loader = PrefetchLoader(cv_loader, self.device)
         self.state.inject_loader(tr_loader)
         # sample weights
         if tr_weights is not None:
@@ -844,25 +858,23 @@ class Trainer(MonitoredMixin):
         self._log_file: str = os.path.join(self.logging_folder, log_name)
         with open(self._log_file, "w"):
             pass
-        step_tqdm_legacy = None
+        step_tqdm = None
         self._epoch_tqdm: Optional[tqdm] = None
         if self.use_tqdm:
             self._epoch_tqdm = tqdm(list(range(self.state.num_epoch)), position=0)
         while self.state.should_train:
             try:
                 self.state.epoch += 1
-                step_tqdm = iter(self.tr_loader)
+                step_iterator = self.tr_loader
                 if self.use_tqdm:
-                    step_tqdm_legacy = step_tqdm = tqdm(
-                        step_tqdm,
+                    step_tqdm = step_iterator = tqdm(
+                        step_iterator,
                         total=len(self.tr_loader),
                         position=1,
                         leave=False,
                     )
-                for (x_batch, y_batch), batch_indices in step_tqdm:
+                for batch, batch_indices in step_iterator:
                     self.state.step += 1
-                    with timing_context(self, "collate batch", enable=self.timing):
-                        batch = self.inference.collate_batch(x_batch, y_batch)
                     with amp_autocast_context(self.use_amp):
                         with timing_context(self, "model.forward", enable=self.timing):
                             forward_results = self.model(
@@ -890,7 +902,7 @@ class Trainer(MonitoredMixin):
                     with timing_context(self, "loss.backward", enable=self.timing):
                         loss = loss_dict["loss"]
                         if self.use_amp:
-                            loss = self.scaler.scale(loss)  # type: ignore
+                            loss = self.grad_scaler.scale(loss)  # type: ignore
                         loss.backward()
                     if self.clip_norm > 0.0:
                         with timing_context(self, "clip_norm_step", enable=self.timing):
@@ -924,8 +936,8 @@ class Trainer(MonitoredMixin):
                 self._epoch_tqdm.total = self.state.num_epoch
                 self._epoch_tqdm.update()
         if self.use_tqdm:
-            if step_tqdm_legacy is not None:
-                step_tqdm_legacy.close()
+            if step_tqdm is not None:
+                step_tqdm.close()
             assert self._epoch_tqdm is not None
             self._epoch_tqdm.close()
         rs = None
