@@ -20,6 +20,7 @@ tensor_tuple_type = Tuple[torch.Tensor, torch.Tensor]
 general_config_type = Optional[Union[str, Dict[str, Any]]]
 predictor_type = Optional[Callable[[int, TabularDataset, np.ndarray], np.ndarray]]
 evaluator_type = Optional[Callable[[List[np.ndarray], List[np.ndarray]], np.ndarray]]
+tensor_batch_type = Tuple[Dict[str, torch.Tensor], Optional[torch.Tensor]]
 
 
 __all__ = [
@@ -31,4 +32,5 @@ __all__ = [
     "general_config_type",
     "predictor_type",
     "evaluator_type",
+    "tensor_batch_type",
 ]
